import tango
from tango import DevState

dev = tango.DeviceProxy("pipeServer/tango/1")
blob = dev.read_pipe("TestPipe")
print("PipeClient.py: ", blob)

blob = (
    "pipeWriteTest0",
    dict(x=5.9, y=15.1, anInt=-169, str="the test", truth=False, state=DevState.FAULT),
)
dev.write_pipe("TestPipe", blob)

float_list = [3.33, 3.34, 3.35, 3.36]
inner_int_list = [11, 12, 13, 14, 15]
inner_inner_data = [
    ("InnerInnerFirstDE", 111),
    ("InnerInnerSecondDE", float_list),
    ("InnerInnerThirdDE", inner_int_list),
]
inner_inner_blob = ("InnerInner", dict(inner_inner_data))
inner_data = [
    ("InnerFirstDE", "Grenoble"),
    ("InnerSecondDE", inner_inner_blob),
    ("InnerThirdDE", True),
]
inner_blob = ("Inner", dict(inner_data))
int_list = [3, 4, 5, 6]
pipe_data = [("1DE", inner_blob), ("2DE", int_list)]
blob = ("PipeWriteTest1", dict(pipe_data))
dev.write_pipe("TestPipe", blob)



# ------------------------------------------------------------------------------
# This file is part of PyTango (http://pytango.rtfd.io)
#
# Copyright 2006-2012 CELLS / ALBA Synchrotron, Bellaterra, Spain
# Copyright 2013-2014 European Synchrotron Radiation Facility, Grenoble, France
#
# Distributed under the terms of the GNU Lesser General Public License,
# either version 3 of the License, or (at your option) any later version.
# See LICENSE.txt for more info.
# ------------------------------------------------------------------------------

"""
This is an internal PyTango module.
"""

__all__ = ("pytango_pprint_init",)

__docformat__ = "restructuredtext"

from tango._tango import (
    StdStringVector,
    StdLongVector,
    CommandInfoList,
    AttributeInfoList,
    AttributeInfoListEx,
    PipeInfoList,
    DeviceDataHistoryList,
    GroupReplyList,
    GroupAttrReplyList,
    GroupCmdReplyList,
    DbData,
    DbDevInfos,
    DbDevExportInfos,
    DbDevImportInfos,
    DbHistoryList,
    LockerInfo,
    DevCommandInfo,
    AttributeDimension,
    CommandInfo,
    PipeInfo,
    DeviceInfo,
    DeviceAttributeConfig,
    AttributeInfo,
    AttributeAlarmInfo,
    ChangeEventInfo,
    PeriodicEventInfo,
    ArchiveEventInfo,
    AttributeEventInfo,
    AttributeInfoEx,
    DeviceAttribute,
    DeviceAttributeHistory,
    DeviceData,
    DeviceDataHistory,
    DevicePipe,
    DbDatum,
    DbDevInfo,
    DbDevImportInfo,
    DbDevExportInfo,
    DbServerInfo,
    GroupReply,
    GroupAttrReply,
    GroupCmdReply,
    DevError,
    EventData,
    AttrConfEventData,
    DataReadyEventData,
    TimeVal,
    DevFailed,
    CmdArgType,
)

from tango.device_server import AttributeAlarm, EventProperties
from tango.device_server import ChangeEventProp, PeriodicEventProp, ArchiveEventProp
from tango.device_server import AttributeConfig, AttributeConfig_2
from tango.device_server import AttributeConfig_3, AttributeConfig_5
import collections.abc


def __inc_param(obj, name):
    ret = not name.startswith("_")
    ret &= name not in ("except_flags",)
    ret &= not isinstance(getattr(obj, name), collections.abc.Callable)
    return ret


def __single_param(obj, param_name, f=repr, fmt="%s = %s"):
    param_value = getattr(obj, param_name)
    if param_name == "data_type":
        param_value = CmdArgType.values.get(param_value, param_value)
    return fmt % (param_name, f(param_value))


def __struct_params_s(obj, separator=", ", f=repr, fmt="%s = %s"):
    """method wrapper for printing all elements of a struct"""
    s = separator.join(
        [__single_param(obj, n, f, fmt) for n in dir(obj) if __inc_param(obj, n)]
    )
    return s


def __struct_params_repr(obj):
    """method wrapper for representing all elements of a struct"""
    return __struct_params_s(obj)


def __struct_params_str(obj, fmt, f=repr):
    """method wrapper for printing all elements of a struct."""
    return __struct_params_s(obj, "\n", f=f, fmt=fmt)


def __repr__Struct(self):
    """repr method for struct"""
    return f"{self.__class__.__name__}({__struct_params_repr(self)})"


def __str__Struct_Helper(self, f=repr):
    """str method for struct"""
    attrs = [n for n in dir(self) if __inc_param(self, n)]
    fmt = attrs and "%%%ds = %%s" % max(map(len, attrs)) or "%s = %s"
    return f"{self.__class__.__name__}[\n{__struct_params_str(self, fmt, f)}]\n"


def __str__Struct(self):
    return __str__Struct_Helper(self, f=repr)


def __str__Struct_extra(self):
    return __str__Struct_Helper(self, f=str)


def __registerSeqStr():
    """helper function to make internal sequences printable"""
    _SeqStr = lambda self: (self and f"[{', '.join(map(repr, self))}]") or "[]"
    _SeqRepr = lambda self: (self and f"[{', '.join(map(repr, self))}]") or "[]"

    seqs = (
        StdStringVector,
        StdLongVector,
        CommandInfoList,
        AttributeInfoList,
        AttributeInfoListEx,
        PipeInfoList,
        DeviceDataHistoryList,
        GroupReplyList,
        GroupAttrReplyList,
        GroupCmdReplyList,
        DbData,
        DbDevInfos,
        DbDevExportInfos,
        DbDevImportInfos,
        DbHistoryList,
    )

    for seq in seqs:
        seq.__str__ = _SeqStr
        seq.__repr__ = _SeqRepr


def __str__DevFailed(self):
    if isinstance(self.args, collections.abc.Sequence):
        seq_str = "\n".join(map(str, self.args))
        return f"DevFailed[\n{seq_str}]"
    return f"DevFailed[{self.args}]"


def __repr__DevFailed(self):
    return f"DevFailed(args = {repr(self.args)})"


def __str__DevError(self):
    desc = self.desc.replace("\n", "\n           ")
    s = f"""DevError[
    desc = {desc}
  origin = {self.origin}
  reason = {self.reason}
severity = {self.severity}]
"""
    return s


def __registerStructStr():
    """helper method to register str and repr methods for structures"""
    structs = (
        LockerInfo,
        DevCommandInfo,
        AttributeDimension,
        CommandInfo,
        DeviceInfo,
        DeviceAttributeConfig,
        AttributeInfo,
        AttributeAlarmInfo,
        ChangeEventInfo,
        PeriodicEventInfo,
        ArchiveEventInfo,
        AttributeEventInfo,
        AttributeInfoEx,
        PipeInfo,
        DeviceAttribute,
        DeviceAttributeHistory,
        DeviceData,
        DeviceDataHistory,
        DevicePipe,
        DbDatum,
        DbDevInfo,
        DbDevImportInfo,
        DbDevExportInfo,
        DbServerInfo,
        GroupReply,
        GroupAttrReply,
        GroupCmdReply,
        DevError,
        EventData,
        AttrConfEventData,
        DataReadyEventData,
        AttributeConfig,
        AttributeConfig_2,
        AttributeConfig_3,
        AttributeConfig_5,
        ChangeEventProp,
        PeriodicEventProp,
        ArchiveEventProp,
        AttributeAlarm,
        EventProperties,
    )

    for struct in structs:
        struct.__str__ = __str__Struct
        struct.__repr__ = __repr__Struct

    # special case for TimeVal: it already has a str representation itself
    TimeVal.__repr__ = __repr__Struct

    # special case for DevFailed: we want a better pretty print
    # also, because it is an Exception it has the message attribute which
    # generates a Deprecation warning in python 2.6
    DevFailed.__str__ = __str__DevFailed
    DevFailed.__repr__ = __repr__DevFailed

    DevError.__str__ = __str__DevError


def pytango_pprint_init(doc=True):
    __registerSeqStr()
    __registerStructStr()



# ------------------------------------------------------------------------------
# This file is part of PyTango (http://pytango.rtfd.io)
#
# Copyright 2006-2012 CELLS / ALBA Synchrotron, Bellaterra, Spain
# Copyright 2013-2014 European Synchrotron Radiation Facility, Grenoble, France
#
# Distributed under the terms of the GNU Lesser General Public License,
# either version 3 of the License, or (at your option) any later version.
# See LICENSE.txt for more info.
# ------------------------------------------------------------------------------

"""
This is an internal PyTango module.
"""


__all__ = ("PipeData",)

__docformat__ = "restructuredtext"

import inspect

from tango._tango import Except, DispLevel, Pipe, PipeWriteType, UserDefaultPipeProp
from tango.utils import is_non_str_seq, is_pure_str


class PipeData:
    """A helper class that contains the same information one of the items in
    DeviceClass.pipe_list but in object form"""

    def __init__(self, name, class_name, pipe_info=None):
        self.class_name = class_name
        self.pipe_name = name
        self.pipe_write = PipeWriteType.PIPE_READ
        self.display_level = DispLevel.OPERATOR
        if name is None:
            self.read_method_name = None
            self.write_method_name = None
            self.is_allowed_name = None
        else:
            self.read_method_name = "read_" + name
            self.write_method_name = "write_" + name
            self.is_allowed_name = "is_" + name + "_allowed"
        self.pipe_class = None
        self.pipe_args = []
        self.pipe_prop = None
        if pipe_info is not None:
            self.from_pipe_info(pipe_info)

    @classmethod
    def from_dict(cls, pipe_dict):
        pipe_dict = dict(pipe_dict)
        name = pipe_dict.pop("name", None)
        class_name = pipe_dict.pop("class_name", None)
        self = cls(name, class_name)
        self.build_from_dict(pipe_dict)
        return self

    def build_from_dict(self, pipe_dict):
        self.display_level = pipe_dict.pop("display_level", DispLevel.OPERATOR)

        is_access_explicit = "access" in pipe_dict
        if is_access_explicit:
            self.pipe_write = pipe_dict.pop("access")
        else:
            # access is defined by which methods were defined
            r_explicit = "fread" in pipe_dict or "fget" in pipe_dict
            w_explicit = "fwrite" in pipe_dict or "fset" in pipe_dict
            if r_explicit and w_explicit:
                self.pipe_write = PipeWriteType.PIPE_READ_WRITE
            else:
                self.pipe_write = PipeWriteType.PIPE_READ

        fread = pipe_dict.pop("fget", pipe_dict.pop("fread", None))
        if fread is not None:
            if is_pure_str(fread):
                self.read_method_name = fread
            elif inspect.isroutine(fread):
                self.read_method_name = fread.__name__
        fwrite = pipe_dict.pop("fset", pipe_dict.pop("fwrite", None))
        if fwrite is not None:
            if is_pure_str(fwrite):
                self.write_method_name = fwrite
            elif inspect.isroutine(fwrite):
                self.write_method_name = fwrite.__name__
        fisallowed = pipe_dict.pop("fisallowed", None)
        if fisallowed is not None:
            if is_pure_str(fisallowed):
                self.is_allowed_name = fisallowed
            elif inspect.isroutine(fisallowed):
                self.is_allowed_name = fisallowed.__name__
        self.pipe_class = pipe_dict.pop("klass", Pipe)
        self.pipe_args.extend((self.pipe_name, self.display_level, self.pipe_write))
        if len(pipe_dict):
            self.pipe_prop = self.__create_user_default_pipe_prop(pipe_dict)
        return self

    def _set_name(self, name):
        old_name = self.pipe_name
        self.pipe_name = name
        self.pipe_args[0] = name
        if old_name is None:
            if self.read_method_name is None:
                self.read_method_name = "read_" + name
            if self.write_method_name is None:
                self.write_method_name = "write_" + name
            if self.is_allowed_name is None:
                self.is_allowed_name = "is_" + name + "_allowed"

    def __throw_exception(self, msg, meth="create_pipe()"):
        Except.throw_exception("PyDs_WrongPipeDefinition", msg, meth)

    def __create_user_default_pipe_prop(self, extra_info):
        """for internal usage only"""
        p = UserDefaultPipeProp()

        doc = extra_info.pop("doc", None)
        if doc is not None:
            extra_info["description"] = doc

        for k, v in extra_info.items():
            k_lower = k.lower()
            method_name = f"set_{k_lower.replace(' ', '_')}"
            if hasattr(p, method_name):
                method = getattr(p, method_name)
                method(str(v))
            else:
                msg = (
                    f"Wrong definition of pipe. "
                    f"The object extra information '{k}' "
                    f"is not recognized!"
                )
                Except.throw_exception(
                    "PyDs_WrongPipeDefinition", msg, "create_user_default_pipe_prop()"
                )
        return p

    def from_pipe_info(self, pipe_info):
        name = self.class_name
        pipe_name = self.pipe_name
        throw_ex = self.__throw_exception
        # check for well defined pipe info

        # check parameter
        if not is_non_str_seq(pipe_info):
            throw_ex(
                f"Wrong data type for value for describing pipe {pipe_name} in "
                f"class {name}\nMust be a sequence with 1 or 2 elements"
            )

        if len(pipe_info) < 1 or len(pipe_info) > 2:
            throw_ex(
                f"Wrong number of argument for describing pipe {pipe_name} in "
                f"class {name}\nMust be a sequence with 1 or 2 elements"
            )

        extra_info = {}
        if len(pipe_info) == 2:
            # pipe_info[1] must be a dictionary
            # extra_info = pipe_info[1], with all the keys lowercase
            for k, v in pipe_info[1].items():
                extra_info[k.lower()] = v

        pipe_info = pipe_info[0]

        # get write type
        try:
            self.pipe_write = PipeWriteType(pipe_info)
        except Exception:
            throw_ex(
                f"Wrong data write type in pipe argument for "
                f"pipe {pipe_name} in class {name}\nPipe write type must be a "
                f"tango.PipeWriteType"
            )
        try:
            self.display_level = DispLevel(
                extra_info.get("display level", DispLevel.OPERATOR)
            )
        except Exception:
            throw_ex(
                f"Wrong display level in pipe information for "
                f"pipe {pipe_name} in class {name}\nPipe information for "
                f"display level is not a tango.DispLevel"
            )

        self.pipe_class = extra_info.get("klass", Pipe)
        self.pipe_args.extend((self.pipe_name, self.display_level, self.pipe_write))

        pipe_prop = None
        if extra_info:
            pipe_prop = self.__create_user_default_pipe_prop(extra_info)
        self.pipe_prop = pipe_prop

    def to_pipe(self):
        pipe = self.pipe_class(*self.pipe_args)
        if self.pipe_prop is not None:
            pipe.set_default_properties(self.pipe_prop)
        return pipe



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Demo power supply tango device server"""

import time
import numpy

from tango import AttrQuality, AttrWriteType, DispLevel, DevState, DebugIt
from tango.server import Device, attribute, command, pipe, device_property


class PowerSupply(Device):
    voltage = attribute(
        label="Voltage",
        dtype=float,
        display_level=DispLevel.OPERATOR,
        access=AttrWriteType.READ,
        unit="V",
        format="8.4f",
        doc="the power supply voltage",
    )

    current = attribute(
        label="Current",
        dtype=float,
        display_level=DispLevel.EXPERT,
        access=AttrWriteType.READ_WRITE,
        unit="A",
        format="8.4f",
        min_value=0.0,
        max_value=8.5,
        min_alarm=0.1,
        max_alarm=8.4,
        min_warning=0.5,
        max_warning=8.0,
        fget="get_current",
        fset="set_current",
        doc="the power supply current",
    )

    noise = attribute(label="Noise", dtype=((int,),), max_dim_x=1024, max_dim_y=1024)

    info = pipe(label="Info")

    host = device_property(dtype=str)
    port = device_property(dtype=int, default_value=9788)

    def init_device(self):
        Device.init_device(self)
        self.__current = 0.0
        self.set_state(DevState.STANDBY)

    def read_voltage(self):
        self.info_stream("read_voltage(%s, %d)", self.host, self.port)
        return 9.99, time.time(), AttrQuality.ATTR_WARNING

    def get_current(self):
        return self.__current

    def set_current(self, current):
        # should set the power supply current
        self.__current = current

    def read_info(self):
        return "Information", dict(
            manufacturer="Tango", model="PS2000", version_number=123
        )

    @DebugIt()
    def read_noise(self):
        return numpy.random.random_integers(1000, size=(100, 100))

    @command
    def TurnOn(self):
        # turn on the actual power supply here
        self.set_state(DevState.ON)

    @command
    def TurnOff(self):
        # turn off the actual power supply here
        self.set_state(DevState.OFF)

    @command(
        dtype_in=float,
        doc_in="Ramp target current",
        dtype_out=bool,
        doc_out="True if ramping went well, False otherwise",
    )
    def Ramp(self, target_current):
        # should do the ramping
        return True


if __name__ == "__main__":
    PowerSupply.run_server()



# ------------------------------------------------------------------------------
# This file is part of PyTango (http://pytango.rtfd.io)
#
# Copyright 2006-2012 CELLS / ALBA Synchrotron, Bellaterra, Spain
# Copyright 2013-2014 European Synchrotron Radiation Facility, Grenoble, France
#
# Distributed under the terms of the GNU Lesser General Public License,
# either version 3 of the License, or (at your option) any later version.
# See LICENSE.txt for more info.
# ------------------------------------------------------------------------------

"""
This is an internal PyTango module.
"""

__all__ = ("exception_init",)

__docformat__ = "restructuredtext"

from tango.utils import document_static_method as __document_static_method
from tango._tango import Except, DevError, ErrSeverity


def __to_dev_failed(exc_type=None, exc_value=None, traceback=None):
    """to_dev_failed(exc_type, exc_value, traceback) -> tango.DevFailed

        Generate a TANGO DevFailed exception.
        The exception is created with a single :class:`~tango.DevError`
        object. A default value *tango.ErrSeverity.ERR* is defined for
        the :class:`~tango.DevError` severity field.

        The parameters are the same as the ones generates by a call to
        :func:`sys.exc_info`.

    Parameters :
        - type : (class)  the exception type of the exception being handled
        - value : (object) exception parameter (its associated value or the
                  second argument to raise, which is always a class instance
                  if the exception type is a class object)
        - traceback : (traceback) traceback object

    Return     : (tango.DevFailed) a tango exception object

    New in PyTango 7.2.1"""
    try:
        Except.throw_python_exception(exc_type, exc_value, traceback)
    except Exception as e:
        return e


# -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~
# DevError pickle
# -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~


def __DevError__getinitargs__(self):
    return ()


def __DevError__getstate__(self):
    return self.reason, self.desc, self.origin, int(self.severity)


def __DevError__setstate__(self, state):
    self.reason = state[0]
    self.desc = state[1]
    self.origin = state[2]
    self.severity = ErrSeverity(state[3])


def __init_DevError():
    DevError.__getinitargs__ = __DevError__getinitargs__
    DevError.__getstate__ = __DevError__getstate__
    DevError.__setstate__ = __DevError__setstate__


def __init_Except():
    Except.to_dev_failed = staticmethod(__to_dev_failed)


def __doc_Except():
    def document_static_method(method_name, desc, append=True):
        return __document_static_method(Except, method_name, desc, append)

    Except.__doc__ = """
    A containner for the static methods:

        - throw_exception
        - re_throw_exception
        - print_exception
        - compare_exception"""

    document_static_method(
        "throw_exception",
        """
    throw_exception(reason, desc, origin, sever=tango.ErrSeverity.ERR) -> None

            Generate and throw a TANGO DevFailed exception.
            The exception is created with a single :class:`~tango.DevError`
            object. A default value *tango.ErrSeverity.ERR* is defined for
            the :class:`~tango.DevError` severity field.

        Parameters :
            - reason : (str) The exception :class:`~tango.DevError` object reason field
            - desc   : (str) The exception :class:`~tango.DevError` object desc field
            - origin : (str) The exception :class:`~tango.DevError` object origin field
            - sever  : (tango.ErrSeverity) The exception DevError object severity field

        Throws     : DevFailed
    """,
    )

    document_static_method(
        "re_throw_exception",
        """
    re_throw_exception(ex, reason, desc, origin, sever=tango.ErrSeverity.ERR) -> None

            Re-throw a TANGO :class:`~tango.DevFailed` exception with one more error.
            The exception is re-thrown with one more :class:`~tango.DevError` object.
            A default value *tango.ErrSeverity.ERR* is defined for the new
            :class:`~tango.DevError` severity field.

        Parameters :
            - ex     : (tango.DevFailed) The :class:`~tango.DevFailed` exception
            - reason : (str) The exception :class:`~tango.DevError` object reason field
            - desc   : (str) The exception :class:`~tango.DevError` object desc field
            - origin : (str) The exception :class:`~tango.DevError` object origin field
            - sever  : (tango.ErrSeverity) The exception DevError object severity field

        Throws     : DevFailed
    """,
    )

    document_static_method(
        "print_error_stack",
        """
    print_error_stack(ex) -> None

            Print all the details of a TANGO error stack.

        Parameters :
            - ex     : (tango.DevErrorList) The error stack reference
    """,
    )

    document_static_method(
        "print_exception",
        """
    print_exception(ex) -> None

            Print all the details of a TANGO exception.

        Parameters :
            - ex     : (tango.DevFailed) The :class:`~tango.DevFailed` exception
    """,
    )

    document_static_method(
        "throw_python_exception",
        """
    throw_python_exception(type, value, traceback) -> None

            Generate and throw a TANGO DevFailed exception.
            The exception is created with a single :class:`~tango.DevError`
            object. A default value *tango.ErrSeverity.ERR* is defined for
            the :class:`~tango.DevError` severity field.

            The parameters are the same as the ones generates by a call to
            :func:`sys.exc_info`.

        Parameters :
            - type : (class)  the exception type of the exception being handled
            - value : (object) exception parameter (its associated value or the
                      second argument to raise, which is always a class instance
                      if the exception type is a class object)
            - traceback : (traceback) traceback object

        Throws     : DevFailed

        New in PyTango 7.2.1
    """,
    )


def __doc_DevError():
    DevError.__doc__ = """
    Structure describing any error resulting from a command execution,
    or an attribute query, with following members:

        - reason : (str) reason
        - severity : (ErrSeverity) error severty (WARN, ERR, PANIC)
        - desc : (str) error description
        - origin : (str) Tango server method in which the error happened"""


def exception_init(doc=True):
    __init_Except()
    __init_DevError()
    if doc:
        __doc_Except()
        __doc_DevError()
